/***************************************************************************
    file	         : kb_basequery.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/



#include	"kb_dom.h"
#include	"kb_dblink.h"
#include	"kb_basequery.h"


/*  mapExpression: Map expression for driver				*/
/*  server	 : KBServer *	   : Server in question			*/
/*  expr	 : const QString & : Expression				*/
/*  (returns)	 : QString	   : Mapped expression			*/

static	QString	mapExpression
	(	KBServer	*server,
		const QString	&expr
	)
{
	return	server == 0 ? expr : server->mapExpression (expr) ;
}

/*  placeHolder	 : Map placeholder for driver				*/
/*  server	 : KBServer *	   : Server in question			*/
/*  place	 : uint		   : Placeholder index			*/
/*  (returns)	 : QString	   : Mapped placeholder			*/

static	QString	placeHolder
	(	KBServer	*server,
		uint		place
	)
{
	return	server == 0 ? "?" : server->placeHolder (place) ;
}

/*  ------------------------------------------------------------------	*/

/*  KBBaseQueryValue							*/
/*  KBBaseQueryValue							*/
/*		: Constructor for base query value			*/
/*  (returns)	: void		:					*/

KBBaseQueryValue::KBBaseQueryValue ()
	:
	m_type	('X')
{
}

/*  KBBaseQueryValue							*/
/*  KBBaseQueryValue							*/
/*		: Constructor for base query value			*/
/*  field	: const QString & : Field name				*/
/*  (returns)	: void		  :					*/

KBBaseQueryValue::KBBaseQueryValue
	(	const QString	&field
	)
	:
	m_field	(field),
	m_type	('V')
{
}

/*  KBBaseQueryValue							*/
/*  KBBaseQueryValue							*/
/*		: Constructor for base query value			*/
/*  field	: const QString & : Field name				*/
/*  text	: const QString & : Field value				*/
/*  (returns)	: void		  :					*/

KBBaseQueryValue::KBBaseQueryValue
	(	const QString	&field,
		const QString	&text
	)
	:
	m_field	(field),
	m_type	('S'),
	m_text	(text)
{
}

/*  KBBaseQueryValue							*/
/*  KBBaseQueryValue							*/
/*		: Constructor for base query value			*/
/*  field	: const QString & : Field name				*/
/*  number	: int		  : Field value				*/
/*  (returns)	: void		  :					*/

KBBaseQueryValue::KBBaseQueryValue
	(	const QString	&field,
		int		number
	)
	:
	m_field	(field),
	m_type	('D'),
	m_fixed	(number)
{
}

/*  KBBaseQueryValue							*/
/*  KBBaseQueryValue							*/
/*		: Constructor for base query value			*/
/*  field	: const QString & : Field name				*/
/*  number	: double	  : Field value				*/
/*  (returns)	: void		  :					*/

KBBaseQueryValue::KBBaseQueryValue
	(	const QString	&field,
		double		number
	)
	:
	m_field	(field),
	m_type	('F'),
	m_float	(number)
{
}

/*  KBBaseQueryValue							*/
/*  KBBaseQueryValue							*/
/*		: Constructor for base query value			*/
/*  field	: const QString & : Field name				*/
/*  		: void *	  : Not used, denotes explicit null	*/
/*  (returns)	: void		  :					*/

KBBaseQueryValue::KBBaseQueryValue
	(	const QString	&field,
		void		*
	)
	:
	m_field	(field),
	m_type	('P')
{
}

/*  KBBaseQueryValue							*/
/*  KBBaseQueryValue							*/
/*		: Constructor for base query value			*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: void		      :					*/

KBBaseQueryValue::KBBaseQueryValue
	(	const QDomElement	&elem
	)
	:
	m_field	(elem.attribute("name")),
	m_type	(elem.attribute("type").at(0).latin1())
{
	QString	value	= elem.attribute ("value") ;

	switch (m_type)
	{
		case 'S' : m_text	= value		   ; break ;
		case 'D' : m_fixed	= value.toInt()    ; break ;
		case 'F' : m_float	= value.toDouble() ; break ;
		default	 :				     break ;
	}
}

/*  KBBaseQueryValue							*/
/*  addToUpdate	: Add query value to update query			*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  place	: uint		: Current place holder index		*/
/*  terms	: QStringList &	: List of terms being assembled		*/
/*  (returns)	: uint		: Possibly incremented place holder	*/

uint	KBBaseQueryValue::addToUpdate
	(	KBServer	*server,
		uint		place,
		QStringList	&terms
	)
	const
{
	QString	expr	;

	switch (m_type)
	{
		case 'V' :
			expr	 = placeHolder(server, place) ;
			place	+= 1 ;
			break	;

		case 'S' :
			expr	= "'" + m_text + "'" ;
			break	;

		case 'D' :
			expr	= QString::number (m_fixed) ;
			break	;

		case 'F' :
			expr	= QString::number (m_float) ;
			break	;

		default	:
			expr	= "null" ;
			break	;
	}

	terms.append
	(	QString("%1 = %3")
			.arg(mapExpression (server, m_field))
			.arg(expr)
	)	;

	return	place	;
}

/*  KBBaseQueryValue							*/
/*  addToInsert	: Add query value to insert query			*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  place	: uint		: Current place holder index		*/
/*  fields	: QStringList &	: List of fields being assembled	*/
/*  exprs	: QStringList &	: List of expressions being assembled	*/
/*  (returns)	: uint		: Possibly incremented place holder	*/

uint	KBBaseQueryValue::addToInsert
	(	KBServer	*server,
		uint		place,
		QStringList	&fields,
		QStringList	&exprs
	)
	const
{
	/* This call is used when building an insert query, so the	*/
	/* field and the expression are added to separate lists.	*/
	fields.append (mapExpression (server, m_field)) ;

	switch (m_type)
	{
		case 'V' :
			exprs.append (placeHolder(server, place)) ;
			place	+= 1 ;
			break	;

		case 'S' :
			exprs.append ("'" + m_text + "'") ;
			break	;

		case 'D' :
			exprs.append (QString::number (m_fixed)) ;
			break	;

		case 'F' :
			exprs.append (QString::number (m_float)) ;
			break	;

		default	:
			exprs.append ("null") ;
			break	;
	}

	return	place	;
}


/*  ------------------------------------------------------------------  */


const char *KBBaseQueryExpr::m_asis	= "asis" ;


/*  KBBaseQueryExpr							*/
/*  KBBaseQueryExpr							*/
/*		: Constructor for base query expression			*/
/*  (returns)	: void		:					*/

KBBaseQueryExpr::KBBaseQueryExpr ()
	:
	m_type	('X')
{
}

/*  KBBaseQueryExpr							*/
/*  KBBaseQueryExpr							*/
/*		: Constructor for base query expression			*/
/*  field	: const QString & : Field name				*/
/*  oper	: const char *	  : Operator				*/
/*  (returns)	: void		  :					*/

KBBaseQueryExpr::KBBaseQueryExpr
	(	const QString	&field,
		const char	*oper
	)
	:
	m_field	(field),
	m_type	(oper == KBBaseQueryExpr::m_asis ? 'A' : 'V'),
	m_oper	(oper)
{
}

/*  KBBaseQueryExpr							*/
/*  KBBaseQueryExpr							*/
/*		: Constructor for base query expression			*/
/*  field	: const QString & : Field name				*/
/*  text	: const QString & : Field value				*/
/*  oper	: const char *    : Operator				*/
/*  (returns)	: void		  :					*/

KBBaseQueryExpr::KBBaseQueryExpr
	(	const QString	&field,
		const QString	&text,
		const char	*oper
	)
	:
	m_field	(field),
	m_type	('S'),
	m_text	(text),
	m_oper	(oper)
{
}

/*  KBBaseQueryExpr							*/
/*  KBBaseQueryExpr							*/
/*		: Constructor for base query expression			*/
/*  field	: const QString & : Field name				*/
/*  number	: int		  : Field value				*/
/*  oper	: const char *    : Operator				*/
/*  (returns)	: void		  :					*/

KBBaseQueryExpr::KBBaseQueryExpr
	(	const QString	&field,
		int		number,
		const char	*oper
	)
	:
	m_field	(field),
	m_type	('D'),
	m_fixed	(number),
	m_oper	(oper)
{
}

/*  KBBaseQueryExpr							*/
/*  KBBaseQueryExpr							*/
/*		: Constructor for base query expression			*/
/*  field	: const QString & : Field name				*/
/*  number	: double	  : Field value				*/
/*  oper	: const char *    : Operator				*/
/*  (returns)	: void		  :					*/

KBBaseQueryExpr::KBBaseQueryExpr
	(	const QString	&field,
		double		number,
		const char	*oper
	)
	:
	m_field	(field),
	m_type	('F'),
	m_float	(number),
	m_oper	(oper)
{
}

/*  KBBaseQueryExpr							*/
/*  KBBaseQueryExpr							*/
/*		: Constructor for base query expression			*/
/*  field	: const QString & : Field name				*/
/*  		: void *	  : Not used, denotes explicit null	*/
/*  oper	: const char *    : Operator				*/
/*  (returns)	: void		  :					*/

KBBaseQueryExpr::KBBaseQueryExpr
	(	const QString	&field,
		void		*,
		const char	*oper
	)
	:
	m_field	(field),
	m_type	('P'),
	m_oper	(oper)
{
}

/*  KBBaseQueryExpr							*/
/*  KBBaseQueryExpr							*/
/*		: Constructor for base query value			*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: void		      :					*/

KBBaseQueryExpr::KBBaseQueryExpr
	(	const QDomElement	&elem
	)
	:
	m_field	(elem.attribute("name")),
	m_type	(elem.attribute("type").at(0).latin1()),
	m_oper	(elem.attribute("oper"))
{
	QString	value	= elem.attribute ("value") ;

	switch (m_type)
	{
		case 'S' : m_text	= value		   ; break ;
		case 'D' : m_fixed	= value.toInt()    ; break ;
		case 'F' : m_float	= value.toDouble() ; break ;
		default	 :				     break ;
	}
}

/*  KBBaseQueryExpr							*/
/*  addToQuery	: Add expression to query				*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  place	: uint		: Current place holder index		*/
/*  terms	: QStringList &	: List of terms being assembled		*/
/*  (returns)	: uint		: Possibly incremented place holder	*/

uint	KBBaseQueryExpr::addToQuery
	(	KBServer	*server,
		uint		place,
		QStringList	&terms
	)
	const
{
	/* This is used for where, group by, etc., hence we form terms	*/
	/* like "a = b", except that for explicit nulls use "a is null"	*/
	QString	expr	;
	QString	oper	= m_oper ;

	if (oper.isEmpty()) oper = "=" ;

	/* Type 'A' (for "asis") is special. In this case "m_field"	*/
	/* contains the complete expression.				*/
	if (m_type == 'A')
	{
		terms.append (mapExpression (server, m_field)) ;
		return	place	;
	}

	switch (m_type)
	{
		case 'V' :
			expr	 = placeHolder(server, place) ;
			place	+= 1 ;
			break	;

		case 'S' :
			expr	= "'" + m_text + "'" ;
			break	;

		case 'D' :
			expr	= QString::number (m_fixed) ;
			break	;

		case 'F' :
			expr	= QString::number (m_float) ;
			break	;

		default	:
			/* Fix the operator. We assume that we are only	*/
			/* called with "=" or "!=".			*/
			expr	= "null" ;
			oper	= oper == "=" ? "is" : "is not" ;
			break	;
	}

	terms.append
	(	QString("%1 %2 %3")
			.arg(mapExpression (server, m_field))
			.arg(oper)
			.arg(expr)
	)	;

	return	place	;
}


/*  ------------------------------------------------------------------  */

/*  KBBaseQueryFetch							*/
/*  KBBaseQueryFetch							*/
/*		: Constructor for expression in select			*/
/*  (returns)	: KBBaseQueryFetch:					*/

KBBaseQueryFetch::KBBaseQueryFetch ()
{
}

/*  KBBaseQueryFetch							*/
/*  KBBaseQueryFetch							*/
/*		: Constructor for expression in select			*/
/*  expr	: const QString & : Expression itself			*/
/*  alias	: const QString & : Possible alias			*/
/*  (returns)	: KBBaseQueryFetch:					*/

KBBaseQueryFetch::KBBaseQueryFetch
	(	const QString	&expr,
		const QString	&alias
	)
	:
	m_expr	(expr),
	m_alias	(alias)
{
}

/*  KBBaseQueryFetch							*/
/*  KBBaseQueryFetch							*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: KBBaseQueryFetch    :					*/

KBBaseQueryFetch::KBBaseQueryFetch
	(	const QDomElement	&elem
	)
	:
	m_expr	(elem.attribute("expr" )),
	m_alias	(elem.attribute("alias"))
{
}

/*  KBBaseQueryFetch							*/
/*  exprText	: Get expression text					*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  (returns)	: QString	: Text					*/

QString	KBBaseQueryFetch::exprText
	(	KBServer	*server
	)
	const
{
	QString	expr	= m_expr	;

	if (server != 0)
		expr	= server->mapExpression (expr) ;

	if (!m_alias.isEmpty())
		return	QString("%1 as %2").arg(expr).arg(m_alias) ;

	return	expr	;
}


/*  ------------------------------------------------------------------  */

/*  KBBaseQueryTable							*/
/*  KBBaseQueryTable							*/
/*		: Constuctor for select query table entry		*/
/*  (returns)	: KBBaseQueryTable :					*/

KBBaseQueryTable::KBBaseQueryTable ()
	:
	m_joinType	(None)
{
}

/*  KBBaseQueryTable							*/
/*  KBBaseQueryTable							*/
/*		: Constuctor for select query table entry		*/
/*  tabName	: const QString &  : Table name				*/
/*  alias	: const QString &  : Table alias			*/
/*  primary	: const QString &  : Primary column			*/
/*  (returns)	: KBBaseQueryTable :					*/

KBBaseQueryTable::KBBaseQueryTable
	(	const QString	&tabName,
		const QString	&alias,
		const QString	&primary
	)
	:
	m_tabName	(tabName),
	m_alias		(alias),
	m_joinType	(None),
	m_primary	(primary)
{
}

/*  KBBaseQueryTable							*/
/*  KBBaseQueryTable							*/
/*		: Constuctor for select query table entry		*/
/*  tabName	: const QString &  : Table name				*/
/*  alias	: const QString &  : Table alias			*/
/*  joinType	: JoinType	   : Join type				*/
/*  joinExpr	: const QString &  : Join expression			*/
/*  primary	: const QString &  : Primary column			*/
/*  (returns)	: KBBaseQueryTable :					*/

KBBaseQueryTable::KBBaseQueryTable
	(	const QString	&tabName,
		const QString	&alias,
		JoinType	joinType,
		const QString	&joinExpr,
		const QString	&primary
	)
	:
	m_tabName	(tabName),
	m_alias		(alias),
	m_joinType	(joinType),
	m_joinExpr	(joinExpr),
	m_primary	(primary)
{
	if (m_joinExpr.isEmpty()) m_joinType = None ;
}

/*  KBBaseQueryTable							*/
/*  KBBaseQueryTable							*/
/*		: Constuctor for select query table entry		*/
/*  tabName	: const QString &  : Table name				*/
/*  alias	: const QString &  : Table alias			*/
/*  joinType	: JoinType	   : Join type				*/
/*  joinExpr	: const QString &  : Join expression			*/
/*  primary	: const QString &  : Primary column			*/
/*  (returns)	: KBBaseQueryTable :					*/

KBBaseQueryTable::KBBaseQueryTable
	(	const QString	&tabName,
		const QString	&alias,
		const QString	&joinType,
		const QString	&joinExpr,
		const QString	&primary
	)
	:
	m_tabName	(tabName),
	m_alias		(alias),
	m_joinType	(joinType == "left"  ? LeftOuter  :
			 joinType == "right" ? RightOuter :
					       Inner 
			),
	m_joinExpr	(joinExpr),
	m_primary	(primary)
{
	if (m_joinExpr.isEmpty()) m_joinType = None ;
}

/*  KBBaseQueryTable							*/
/*  KBBaseQueryTable							*/
/*		: Constuctor for select query table entry		*/
/*  elem	: const QDomElement & : Definition			*/
/*  tabName	: const QString &  : Table name				*/
/*  alias	: const QString &  : Table alias			*/
/*  joinType	: JoinType	   : Join type				*/
/*  joinExpr	: const QString &  : Join expression			*/
/*  primary	: const QString &  : Primary column			*/
/*  (returns)	: KBBaseQueryTable    :					*/

KBBaseQueryTable::KBBaseQueryTable
	(	const QDomElement	&elem
	)
	:
	m_tabName	(elem.attribute("name"   )),
	m_alias		(elem.attribute("alias"  )),
	m_joinType	(
			 elem.attribute("jtype"  ) == "left"  ? LeftOuter  :
			 elem.attribute("jtype"  ) == "right" ? RightOuter :
					       		        Inner 
			),
	m_joinExpr	(elem.attribute("jexpr"  )),
	m_primary	(elem.attribute("primary"))
{
	if (m_joinExpr.isEmpty()) m_joinType = None ;
}

/*  KBBaseQueryTable							*/
/*  hasOuterJoin : See if table has an outer join			*/
/*  (returns)	 : bool		: True if so				*/

bool	KBBaseQueryTable::hasOuterJoin () const
{
	return	(m_joinType == RightOuter) || (m_joinType == LeftOuter) ;
}

/*  KBBaseQueryTable							*/
/*  hasInnerJoin : See if table has an inner join			*/
/*  (returns)	 : bool		: True if so				*/

bool	KBBaseQueryTable::hasInnerJoin () const
{
	return	(m_joinType == Inner) ;
}

/*  KBBaseQueryTable							*/
/*  hasAnyJoin	 : See if table has an any join				*/
/*  (returns)	 : bool		: True if so				*/

bool	KBBaseQueryTable::hasAnyJoin () const
{
	return	(m_joinType != None) ;
}

/*  KBBaseQueryTable							*/
/*  tableText	: Get text for table with possible alias		*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  (returns)	: QString	: Text					*/

QString	KBBaseQueryTable::tableText
	(	KBServer	*server
	)
	const
{
	QString	tabName	= m_tabName	;

	if (server != 0)
		tabName	= server->mapExpression (tabName) ;

	if (!m_alias.isEmpty())
		return	QString("%1 %2").arg(tabName).arg(m_alias) ;

	return	tabName ;
}

/*  KBBaseQueryTable							*/
/*  joinType	 : Get text for table join type				*/
/*  (returns)	 : QString	: Text					*/

QString	KBBaseQueryTable::joinType () const
{
	switch (m_joinType)
	{
		case LeftOuter  : return "left outer join"  ;
		case RightOuter : return "right outer join" ;
		case Inner	: return "inner join"	    ;
		default	   	: break  ;
	}

	return	"unknown join" ;
}

/*  KBBaseQueryTable							*/
/*  joinExpr	: Get text for table join expression			*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  (returns)	: QString	: Text					*/

QString	KBBaseQueryTable::joinExpr
	(	KBServer	*server
	)
	const
{
	return	server == 0 ? m_joinExpr : server->mapExpression (m_joinExpr) ;
}

/*  KBBaseQueryTable							*/
/*  tableName	: Return table name					*/
/*  (returns)	: const QString & : Table name				*/

const QString &KBBaseQueryTable::tableName () const
{
	return	m_tabName ;
}

/*  ------------------------------------------------------------------	*/

/*  KBBaseQuery								*/
/*  KBBaseQuery	: Query base class constructor				*/
/*  table	: const QString & : Optional initial table		*/
/*  (returns)	: KBBaseQuery	  :					*/

KBBaseQuery::KBBaseQuery
	(	const QString		&table
	)
{
	
	if (!table.isEmpty()) setTable (table) ;
}

/*  KBBaseQuery								*/
/*  KBBaseQuery	: Query base class constructor				*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: KBBaseQuery	      :					*/

KBBaseQuery::KBBaseQuery
	(	const QDomElement	&elem
	)
{
	DOMITER_ELEM_BEGIN(elem,child)
	{
		if (child.tagName() == "table")
		{
			setTable (child.attribute("name")) ;
			continue ;
		}
		if (child.tagName() == "value")
		{
			addValue (child) ;
			continue ;
		}
		if (child.tagName() == "where")
		{
			addWhere (child) ;
			continue ;
		}
	}
	DOMITER_END
}

/*  KBBaseQuery								*/
/*  ~KBBaseQuery: Query base class constructor				*/
/*  (returns)	:		  :					*/

KBBaseQuery::~KBBaseQuery ()
{
}

/*  KBBaseQuery								*/
/*  setTable	: Set a single table name				*/
/*  table	: const QString & : Table name				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::setTable
	(	const QString	&table
	)
{
	m_tableList.clear  () ;
	m_tableList.append (KBBaseQueryTable (table, QString::null)) ;
}

/*  KBBaseQuery								*/
/*  addValue	: Add value for update and insert queries		*/
/*  field	: const QString & : Field name				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addValue
	(	const QString 	&field
	)
{
	/* This version is used when the value will be passed when the	*/
	/* query is executed, and implies a placeholder.		*/
	m_valueList.append (KBBaseQueryValue (field)) ;
}

/*  KBBaseQuery								*/
/*  addValue	: Add value for update and insert queries		*/
/*  field	: const QString & : Field name				*/
/*  text	: const QString & : Text value				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addValue
	(	const QString 	&field,
		const QString 	&text
	)
{
	m_valueList.append (KBBaseQueryValue (field, text)) ;
}

/*  KBBaseQuery								*/
/*  addValue	: Add value for update and insert queries		*/
/*  field	: const QString & : Field name				*/
/*  number	: int		  : Integer value			*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addValue
	(	const QString 	&field,
		int   		number
	)
{
	m_valueList.append (KBBaseQueryValue (field, number)) ;
}

/*  KBBaseQuery								*/
/*  addValue	: Add value for update and insert queries		*/
/*  field	: const QString & : Field name				*/
/*  number	: double	  : Floating value			*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addValue
	(	const QString 	&field,
		double		number
	)
{
	m_valueList.append (KBBaseQueryValue (field, number)) ;
}

/*  KBBaseQuery								*/
/*  addValue	: Add value for update and insert queries		*/
/*  field	: const QString & : Field name				*/
/*  p		: void *	  : Not used, implies a null value	*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addValue
	(	const QString 	&field,
		void 		*p
	)
{
	m_valueList.append (KBBaseQueryValue (field, p)) ;
}

/*  KBBaseQuery								*/
/*  addValue	: Add value for update and insert queries		*/
/*  elem	: const QDomElemment & : Value definition		*/
/*  (returns)	: void		       :					*/

void	KBBaseQuery::addValue
	(	const QDomElement	&elem
	)
{
	m_valueList.append (KBBaseQueryValue (elem)) ;
}


/*  KBBaseQuery								*/
/*  addWhere	: Add where condition for select, update and delete	*/
/*  field	: const QString & : Field name				*/
/*  oper	: const char *	  : Operator				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addWhere
	(	const QString 	&field,
		const char	*oper
	)
{
	/* This version is used when the value will be passed when the	*/
	/* query is executed, and implies a placeholder.		*/
	m_whereList.append (KBBaseQueryExpr (field, oper)) ;
}

/*  KBBaseQuery								*/
/*  addWhere	: Add where condition for select, update and delete	*/
/*  field	: const QString & : Field name				*/
/*  text	: const QString & : Text value				*/
/*  oper	: const char *	  : Operator				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addWhere
	(	const QString 	&field,
		const QString 	&text,
		const char	*oper
	)
{
	m_whereList.append (KBBaseQueryExpr (field, text, oper)) ;
}

/*  KBBaseQuery								*/
/*  addWhere	: Add where condition for select, update and delete	*/
/*  field	: const QString & : Field name				*/
/*  number	: int		  : Integer value			*/
/*  oper	: const char *	  : Operator				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addWhere
	(	const QString 	&field,
		int   		number,
		const char	*oper
	)
{
	m_whereList.append (KBBaseQueryExpr (field, number, oper)) ;
}

/*  KBBaseQuery								*/
/*  addWhere	: Add where condition for select, update and delete	*/
/*  field	: const QString & : Field name				*/
/*  number	: double	  : Floating value			*/
/*  oper	: const char *	  : Operator				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addWhere
	(	const QString 	&field,
		double		number,
		const char	*oper
	)
{
	m_whereList.append (KBBaseQueryExpr (field, number, oper)) ;
}

/*  KBBaseQuery								*/
/*  addWhere	: Add where condition for select, update and delete	*/
/*  field	: const QString & : Field name				*/
/*  p		: void *	  : Not used, implies a null value	*/
/*  oper	: const char *	  : Operator				*/
/*  (returns)	: void		  :					*/

void	KBBaseQuery::addWhere
	(	const QString 	&field,
		void 		*p,
		const char	*oper
	)
{
	m_whereList.append (KBBaseQueryExpr (field, p, oper)) ;
}

/*  KBBaseQuery								*/
/*  addWhere	: Add where condition for select, update and delete	*/
/*  elem	: const QDomElemment & : Value definition		*/
/*  (returns)	: void		       :					*/

void	KBBaseQuery::addWhere
	(	const QDomElement	&elem
	)
{
	m_whereList.append (KBBaseQueryExpr (elem)) ;
}


/*  KBBaseQuery								*/
/*  tableName	: Get single table name					*/
/*  (returns)	: QString	: Table name				*/

QString	KBBaseQuery::tableName () const
{
	/* Should only be called for single-table queries so return	*/
	/* null otherwise.						*/
	return	m_tableList.count() == 1 ?
			m_tableList[0].tableName() :
			QString::null		   ;
}

/*  KBBaseQuery								*/
/*  getQueryText: Get text of sql query					*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  (returns)	: QString	: SQL query				*/

QString	KBBaseQuery::getQueryText
	(	KBServer	*server
	)
	const
{
	return	makeQueryText (server) ;
}

/*  KBBaseQuery								*/
/*  getQueryText: Get text of sql query					*/
/*  dbLink	: KBDBLink *	: Database link				*/
/*  (returns)	: QString	: SQL query				*/

QString	KBBaseQuery::getQueryText
	(	KBDBLink	*dbLink
	)
	const
{
	KBServer *server = dbLink == 0 ? 0 : dbLink->server() ;
	return	 makeQueryText (server) ;
}

/*  ------------------------------------------------------------------	*/

/*  KBBaseUpdate							*/
/*  KBBaseUpdate: Constructor for update query object			*/
/*  table	: const QString & : Optional initial table		*/
/*  (returns)	: KBBaseUpdate	  :					*/

KBBaseUpdate::KBBaseUpdate
	(	const QString	&table
	)
	:
	KBBaseQuery (table)
{
}

/*  KBBaseUpdate								*/
/*  KBBaseUpdate: Query base class constructor				*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: KBBaseUpdate	      :					*/

KBBaseUpdate::KBBaseUpdate
	(	const QDomElement	&elem
	)
	:
	KBBaseQuery (elem)
{
}

/*  KBBaseUpdate							*/
/*  makeQueryText							*/
/*		: Get text of sql query					*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  (returns)	: QString	: SQL query				*/

QString	KBBaseUpdate::makeQueryText
	(	KBServer	*server
	)
	const
{
	uint		place	= 0 ;
	QStringList	fields	;
	QStringList	where	;

	for (uint idx1 = 0 ; idx1 < m_valueList.count() ; idx1 += 1)
		place	= m_valueList[idx1].addToUpdate (server, place, fields) ;
	for (uint idx2 = 0 ; idx2 < m_whereList .count() ; idx2 += 1)
		place	= m_whereList[idx2].addToQuery  (server, place, where ) ;

	QString	table	= m_tableList[0].tableName() ;
	if (server != 0) table = server->mapExpression (table) ;

	QString	update	= QString("update %1 set %2")
				 .arg(table)
				 .arg(fields.join(", "))
			  ;

	if (where.count() > 0)
		update += " where " + where.join(" and ") ;

	return	update	;
}

/*  ------------------------------------------------------------------	*/

/*  KBBaseInsert							*/
/*  KBBaseInsert: Constructor for insert query object			*/
/*  table	: const QString & : Optional initial table		*/
/*  (returns)	: KBBaseInsert	  :					*/

KBBaseInsert::KBBaseInsert
	(	const QString	&table
	)
	:
	KBBaseQuery (table)
{
}

/*  KBBaseInsert								*/
/*  KBBaseInsert: Query base class constructor				*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: KBBaseInsert	      :					*/

KBBaseInsert::KBBaseInsert
	(	const QDomElement	&elem
	)
	:
	KBBaseQuery (elem)
{
}

/*  KBBaseInsert							*/
/*  makeQueryText							*/
/*		: Get text of sql query					*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  (returns)	: QString	: SQL query				*/

QString	KBBaseInsert::makeQueryText
	(	KBServer	*server
	)
	const
{
	uint		place	= 0 ;
	QStringList	fields	;
	QStringList	exprs	;

	for (uint idx = 0 ; idx < m_valueList.count() ; idx += 1)
		place	= m_valueList[idx].addToInsert (server, place, fields, exprs) ;

	QString	table	= m_tableList[0].tableName() ;
	if (server != 0) table = server->mapExpression (table) ;

	return	QString("insert into %1 (%2) values (%3)")
			.arg(table)
			.arg(fields.join(", "))
			.arg(exprs .join(", "))
			;
}

/*  ------------------------------------------------------------------	*/

/*  KBBaseSelect							*/
/*  KBBaseSelect: Constructor for base select class			*/
/*  table	: const QString & : Optional table name			*/
/*  (returns)	: KBBaseSelect	  :					*/

KBBaseSelect::KBBaseSelect
	(	const QString		&table
	)
	:
	KBBaseQuery (table)
{
	m_distinct	= false	;
	m_forUpdate	= false	;
	m_offset	= -1	;
	m_limit		= -1	;
}

/*  KBBaseSelect								*/
/*  KBBaseSelect: Query base class constructor				*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: KBBaseSelect	      :					*/

KBBaseSelect::KBBaseSelect
	(	const QDomElement	&elem
	)
	:
	KBBaseQuery (elem)
{
	m_distinct	= elem.attribute("distinct"    ).toInt() ;
	m_forUpdate	= elem.attribute("update"      ).toInt() ;
	m_offset	= elem.attribute("offset", "-1").toInt() ;
	m_limit		= elem.attribute("limit",  "-1").toInt() ;

	/* Clear the table list, we will be loading it again here to	*/
	/* pick up possibly multiple tables.				*/
	m_tableList.clear () ;

	DOMITER_ELEM_BEGIN(elem,child)
	{
		if (child.tagName() == "table")
		{
			addTable (child) ;
			continue ;
		}
		if (child.tagName() == "fetch" )
		{
			addFetch (child) ;
			continue ;
		}
		if (child.tagName() == "group" )
		{
			addGroup (child) ;
			continue ;
		}
		if (child.tagName() == "having")
		{
			addHaving(child) ;
			continue ;
		}
		if (child.tagName() == "order" )
		{
			addOrder (child) ;
			continue ;
		}
	}
	DOMITER_END
}

/*  KBBaseSelect								*/
/*  addTable	: Add a table to the select				*/
/*  tabName	: const QString & : Table name				*/
/*  alias	: const QString & : Table alias				*/
/*  (returns)	: KBBaseSelectTable   :					*/

void	KBBaseSelect::addTable
	(	const QString	&tabName,
		const QString	&alias
	)
{
	m_tableList.append
	(	KBBaseQueryTable
		(	tabName,
			alias
	)	)	;
}

/*  KBBaseSelect								*/
/*  addTable	: Add a table to the select				*/
/*  tabName	: const QString & : Table name				*/
/*  alias	: const QString & : Table alias				*/
/*  joinType	: JoinType	  : Join type				*/
/*  joinExpr	: const QString & : Join expression			*/
/*  (returns)	: KBBaseQueryTable   :					*/

void	KBBaseSelect::addTable
	(	const QString		&tabName,
		const QString		&alias,
		KBBaseQueryTable::JoinType	joinType,
		const QString		&joinExpr
	)
{
	m_tableList.append
	(	KBBaseQueryTable
		(	tabName,
			alias,
			joinType,
			joinExpr
	)	)	;
}

/*  KBBaseSelect								*/
/*  addTable	: Add a table to the select				*/
/*  tabName	: const QString & : Table name				*/
/*  alias	: const QString & : Table alias				*/
/*  joinType	: const QString & : Join type				*/
/*  joinExpr	: const QString & : Join expression			*/
/*  (returns)	: KBBaseQueryTable   :					*/

void	KBBaseSelect::addTable
	(	const QString	&tabName,
		const QString	&alias,
		const QString	&joinType,
		const QString	&joinExpr
	)
{
	m_tableList.append
	(	KBBaseQueryTable
		(	tabName,
			alias,
			joinType,
			joinExpr
	)	)	;
}

/*  KBBaseSelect								*/
/*  addTable	: Add a table to the select					*/
/*  elem	: const QDomElemment & : Table definition			*/
/*  (returns)	: void		       :					*/

void	KBBaseSelect::addTable
	(	const QDomElement	&elem
	)
{
	m_tableList.append (KBBaseQueryTable (elem)) ;
}


/*  KBBaseSelect								*/
/*  addFetch	: Add an expression to the select			*/
/*  expr	: const QString & : Expression				*/
/*  alias	: const QString & : Possible alias			*/
/*  (returns)	: void		  :					*/

void	KBBaseSelect::addFetch
	(	const QString	&expr,
		const QString	&alias
	)
{
	m_fetchList.append (KBBaseQueryFetch (expr, alias)) ;
}

/*  KBBaseSelect								*/
/*  addGroup	: Add a group expression to the select			*/
/*  group	: const QString & : Expression				*/
/*  (returns)	: void		  :					*/

void	KBBaseSelect::addGroup
	(	const QString	&group
	)
{
	m_groupList.append (KBBaseQueryExpr (group, KBBaseQueryExpr::m_asis)) ;
}

/*  KBBaseSelect								*/
/*  addHaving	: Add a having expression to the select			*/
/*  having	: const QString & : Expression				*/
/*  (returns)	: void		  :					*/

void	KBBaseSelect::addHaving
	(	const QString	&having
	)
{
	m_havingList.append (KBBaseQueryExpr (having, KBBaseQueryExpr::m_asis)) ;
}

/*  KBBaseSelect							*/
/*  addOrder	: Add an order expression to the select			*/
/*  order	: const QString & : Expression				*/
/*  (returns)	: void		  :					*/

void	KBBaseSelect::addOrder
	(	const QString	&order
	)
{
	m_orderList.append (KBBaseQueryExpr (order, KBBaseQueryExpr::m_asis)) ;
}


/*  KBBaseSelect							*/
/*  addFetch	: Add an expression to the select			*/
/*  elem	: const QDomElemment & : Table definition		*/
/*  (returns)	: void		       :				*/

void	KBBaseSelect::addFetch
	(	const QDomElement	&elem
	)
{
	m_fetchList.append (KBBaseQueryFetch (elem)) ;
}

/*  KBBaseSelect								*/
/*  addGroup	: Add a group expression to the select			*/
/*  elem	: const QDomElemment & : Table definition		*/
/*  (returns)	: void		  :					*/

void	KBBaseSelect::addGroup
	(	const QDomElement	&elem
	)
{
	m_groupList.append (KBBaseQueryExpr (elem)) ;
}

/*  KBBaseSelect								*/
/*  addHaving	: Add a having expression to the select			*/
/*  elem	: const QDomElemment & : Table definition		*/
/*  (returns)	: void		  :					*/

void	KBBaseSelect::addHaving
	(	const QDomElement	&elem
	)
{
	m_havingList.append (KBBaseQueryExpr (elem)) ;
}

/*  KBBaseSelect							*/
/*  addOrder	: Add an order expression to the select			*/
/*  elem	: const QDomElemment & : Table definition		*/
/*  (returns)	: void		  :					*/

void	KBBaseSelect::addOrder
	(	const QDomElement	&elem
	)
{
	m_orderList.append (KBBaseQueryExpr (elem)) ;
}



/*  KBBaseSelect							*/
/*  setDistinct	: Set query "distinct"					*/
/*  distinct	: bool		: Set of clear				*/
/*  (returns)	: void		:					*/

void	KBBaseSelect::setDistinct
	(	bool		distinct
	)
{
	m_distinct = distinct ;
}

/*  KBBaseSelect							*/
/*  setForUpdate: Set as "select .... for update"			*/
/*  forUpdate	: bool		: Set or clear				*/
/*  (returns)	: void		:					*/

void	KBBaseSelect::setForUpdate
	(	bool		forUpdate
	)
{
	m_forUpdate = forUpdate ;
}

/*  KBBaseSelect							*/
/*  setLimit	: Set query limit and offset values			*/
/*  limit	: int		: Required row limit			*/
/*  offset	: int		: Required forst row offset		*/
/*  (returns)	: void		:					*/

void	KBBaseSelect::setLimit
	(	int		limit,
		int		offset
	)
{
	m_limit	 = limit	;
	m_offset = offset	;
}

/*  KBBaseSelect								*/
/*  makeQueryText							*/
/*		: Get text of sql query					*/
/*  server 	: KBServer *	: Server being accessed			*/
/*  (returns)	: QString	: Text					*/

QString	KBBaseSelect::makeQueryText
	(	KBServer	*server
	)
	const
{
	QString	SQL		;
	cchar	*sep		= ""	;
	bool	hasOuterJoin	= false	;
	uint	place		= 0	;

	for (uint idx1 = 0 ; idx1 < m_tableList.count() ; idx1 += 1)
		if (m_tableList[idx1].hasOuterJoin())
		{	hasOuterJoin = true ;
			break	;
		}

	SQL	= "select " ;
	if (m_distinct) SQL += "distinct " ;

	QStringList fetchList ;
	for (uint idx2 = 0 ; idx2 < m_fetchList.count() ; idx2 += 1)
		fetchList.append (m_fetchList[idx2].exprText(server)) ;
	SQL	+= fetchList.join (", ") ;

	SQL	+= " from " ;

	if (hasOuterJoin)
	{
		QStringList tables3 ;
		for (uint idx3 = 0 ; idx3 < m_tableList.count() ; idx3 += 1)
			if (!m_tableList[idx3].hasAnyJoin())
				tables3.append (m_tableList[idx3].tableText(server)) ;
		SQL	+= tables3.join (", ") ;

		QStringList tables4 ;
		for (uint idx4 = 0 ; idx4 < m_tableList.count() ; idx4 += 1)
			if (m_tableList[idx4].hasAnyJoin())
				tables4.append
				(	QString("%1 %2 on %3")
						.arg(m_tableList[idx4].joinType ()      )
						.arg(m_tableList[idx4].tableText(server))
						.arg(m_tableList[idx4].joinExpr (server))
				)	;
		SQL	+= tables4.join (", ") ;

		sep	= " where " ;
	}
	else
	{
		QStringList tables5 ;
		for (uint idx5 = 0 ; idx5 < m_tableList.count() ; idx5 += 1)
			tables5.append (m_tableList[idx5].tableText(server)) ;
		SQL	+= tables5.join (", ") ;

		QStringList tables6 ;
		for (uint idx6 = 0 ; idx6 < m_tableList.count() ; idx6 += 1)
			if (!m_tableList[idx6].joinExpr(0).isEmpty())
				tables6.append (m_tableList[idx6].joinExpr(server)) ;
		if (tables6.count() > 0)
		{
			SQL	+= " where " + tables6.join (" and ") ;
			sep	 = " and "   ;
		}
		else	sep	 = " where " ;
	}

	if (m_whereList .count() > 0)
	{
		QStringList where  ;
		for (uint idx7 = 0 ; idx7 < m_whereList .count() ; idx7 += 1)
			place = m_whereList [idx7].addToQuery (server, place,  where) ;
		SQL	+= sep          + where.join (" and ") ;
	}
	if (m_groupList .count() > 0)
	{
		QStringList group  ;
		for (uint idx8 = 0 ; idx8 < m_groupList .count() ; idx8 += 1)
			place = m_groupList [idx8].addToQuery (server, place,  group) ;
		SQL	+= " group by " + group.join (" and ") ;

	}
	if (m_havingList.count() > 0)
	{
		QStringList having ;
		for (uint idx9 = 0 ; idx9 < m_havingList.count() ; idx9 += 1)
			place = m_havingList[idx9].addToQuery (server, place, having) ;
		SQL	+= " having "   + having.join (", ") ;
	}
	if (m_orderList .count() > 0)
	{
		QStringList order  ;
		for (uint idxA = 0 ; idxA < m_orderList .count() ; idxA += 1)
			place = m_orderList [idxA].addToQuery (server, place, order ) ;
		SQL	+= " order by " + order.join (", ") ;
	}

	if (m_forUpdate)
		SQL	+= " for update" ;

	if (m_limit > 0)
	{
		if (server != 0)
		{
			QString	text	;
			if (server->getSyntax (text, KBServer::Limit, m_limit, 0))
				SQL	+= text	;
			else	server->lastError().DISPLAY() ;
		}
		else	SQL	+= QString(" limit %1 offset %2").arg(m_limit).arg(0) ;
	}

	return	SQL	;
}

bool	KBBaseSelect::forUpdate () const
{
	return	m_forUpdate ;
}

/*  ------------------------------------------------------------------	*/

KBBaseDelete::KBBaseDelete
	(	const QString	&table
	)
	:
	KBBaseQuery (table)
{
}

/*  KBBaseDelete								*/
/*  KBBaseDelete: Query base class constructor				*/
/*  elem	: const QDomElement & : Definition			*/
/*  (returns)	: KBBaseDelete	      :					*/

KBBaseDelete::KBBaseDelete
	(	const QDomElement	&elem
	)
	:
	KBBaseQuery (elem)
{
}

/*  KBBaseDelete							*/
/*  getQueryText: Get text of sql query					*/
/*  makeQueryText							*/
/*		: Get text of sql query					*/
/*  (returns)	: QString	: SQL query				*/

QString	KBBaseDelete::makeQueryText
	(	KBServer	*server
	)
	const
{
	uint		place	= 0 ;
	QStringList	where	;

	for (uint idx2 = 0 ; idx2 < m_whereList .count() ; idx2 += 1)
		place	= m_whereList[idx2].addToQuery  (server, place, where ) ;

	QString	table	= m_tableList[0].tableName() ;
	if (server != 0) table = server->mapExpression (table) ;

	QString	update	= QString("delete from %1").arg(table) ;

	if (where.count() > 0)
		update += " where " + where.join(" and ") ;

	return	update	;
}
